from typing import Any, Callable, Optional

from django.db.models import Model, QuerySet
from django.forms import BaseModelFormSet, ModelForm, modelformset_factory


class NonrelatedInlineModelFormSet(BaseModelFormSet):
    def __init__(
        self,
        instance: Optional[Model] = None,
        save_as_new: bool = False,
        **kwargs: Any,
    ) -> None:
        self.instance = instance
        super().__init__(**kwargs)
        self.queryset = self.provided_queryset

    @classmethod
    def get_default_prefix(cls: BaseModelFormSet) -> str:
        return f"{cls.model._meta.app_label}-{cls.model._meta.model_name}"

    def save_new(self, form: ModelForm, commit: bool = True):
        obj = super().save_new(form, commit=False)
        self.save_new_instance(self.instance, obj)

        if commit:
            obj.save()

        return obj


def nonrelated_inline_formset_factory(
    model: Model,
    queryset: Optional[QuerySet] = None,
    formset: BaseModelFormSet = NonrelatedInlineModelFormSet,
    save_new_instance: Optional[Callable] = None,
    **kwargs: Any,
) -> BaseModelFormSet:
    inline_formset = modelformset_factory(model, formset=formset, **kwargs)
    inline_formset.provided_queryset = queryset
    inline_formset.save_new_instance = save_new_instance
    return inline_formset



from django.core.paginator import Paginator
from django.utils.functional import cached_property


class InfinitePaginator(Paginator):
    template_name = "unfold/helpers/pagination_infinite.html"

    @cached_property
    def count(self):
        return 9_999_999_999



from collections.abc import Iterable
from typing import Any, Optional, Protocol, Union


class ActionFunction(Protocol):
    """
    Type that specifies functions that are used as actions (annotated with @actions decorator)
    These are functions (__call__) and have 4 additional attributes
    (allowed_permissions, short_description, url_path and attrs)
    """

    allowed_permissions: Iterable[str]
    short_description: str
    url_path: str
    attrs: dict[str, Any]
    icon: Optional[str] = None

    def __call__(self, *args, **kwargs):
        pass


FieldsetsType = Union[
    list[tuple[Union[str, None], dict[str, Any]]],
    tuple[tuple[Union[str, None], dict[str, Any]]],
]



from crispy_forms.layout import BaseInput

from unfold.widgets import BUTTON_CLASSES


class ButtonClassesMixin:
    def __init__(self, *args, css_class=None, **kwargs):
        classes = BUTTON_CLASSES

        if css_class:
            classes.append(css_class)

        self.field_classes = " ".join(classes)

        super().__init__(*args, **kwargs)


class Submit(ButtonClassesMixin, BaseInput):
    input_type = "submit"


class Button(ButtonClassesMixin, BaseInput):
    input_type = "button"
